/* xosdd.c - XOSD proxy daemon
 *
 * Copyright 2009, 2015, 2016, 2022, 2023 Petteri Hintsanen
 * <petterih@iki.fi>
 *
 * This file is part of xosdd.
 *
 * xosdd is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * xosdd is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xosdd.  If not, see <http://www.gnu.org/licenses/>.
 */

#define VERSION "0.1"
#define COPYRIGHT "Copyright 2009, 2015, 2016, 2022, 2023 Petteri Hintsanen"

#define _POSIX_C_SOURCE 2

#include "osd.h"
#include "util.h"
#include <assert.h>
#include <errno.h>
#include <locale.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

static void cleanup_fifo(void);
static void cleanup_osd(void);
static void handle_command(char **tokens);
static void init(void);
static osd *init_osd(void);
static int make_fifo(const char *fifopath);
static void parse_options(int argc, char *argv[]);
static void print_usage(void);
static void print_version(void);
static void process_requests(char *fifopath, osd *osd);
static void set_signal_handlers(void);
static void termination_handler(int signum);

/* structure for holding options and arguments */
/* default values are assigned in parse_options */
struct arguments {
    char *fifo;
    char *align;
    int barlen;
    char *color;
    char *outcolor;
    char *font;
    int horizoff;
    int nlines;
    int outoff;
    char *pos;
    int shadowoff;
    int timeout;
    int vertoff;
} arguments;

static volatile sig_atomic_t cleanup_in_progress = 0;
static int fifo_was_created;
static osd *osd_handle = NULL;

/* recognized commands */
static const char *CMD_ALIGN = "align";
static const char *CMD_BAR = "bar";
static const char *CMD_BAR_LENGTH = "barlen";
static const char *CMD_COLOR = "color";
static const char *CMD_FONT = "font";
static const char *CMD_HORIZONTAL_OFFSET = "horizoff";
static const char *CMD_HIDE = "hide";
static const char *CMD_OUTLINE_COLOR = "outcolor";
static const char *CMD_OUTLINE_OFFSET = "outoff";
static const char *CMD_POSITION = "pos";
static const char *CMD_SCROLL = "scroll";
static const char *CMD_SHADOW_OFFSET = "shadowoff";
static const char *CMD_SHOW = "show";
static const char *CMD_SLIDER = "slider";
static const char *CMD_STR_BAR = "strbar";
static const char *CMD_STRING = "string";
static const char *CMD_STRINGS = "strings";
static const char *CMD_STR_SLIDER = "strslider";
static const char *CMD_TIMEOUT = "timeout";
static const char *CMD_VERTICAL_OFFSET = "vertoff";

/* Clean up fifo, if it was created by us. */
void
cleanup_fifo(void)
{
    if (fifo_was_created) {
        ph_dprintf("removing fifo");
        remove(arguments.fifo);
    }
}

/* Clean up OSD. */
void
cleanup_osd(void)
{
    if (osd_handle) {
        ph_dprintf("destroying OSD");
        destroy_osd(osd_handle);
        osd_handle = NULL;
    }
}

/* Parse a command vector and dispatch it to the osd.  tokens is
 * expected to be an array of pointers to strings, terminated by a
 * null pointer.  The first token should specify a valid command, and
 * the rest of the strings are converted to arguments for that
 * command. */
static void
handle_command(char **tokens)
{
    assert(tokens[0]);

    if (strcmp(tokens[0], CMD_ALIGN) == 0) {
        if (!tokens[1]) {
            ph_eprintf("alignment argument is null");
            return;
        } else {
            set_align(osd_handle, tokens[1]);
        }
    } else if (strcmp(tokens[0], CMD_STR_BAR) == 0
               || strcmp(tokens[0], CMD_STR_SLIDER) == 0) {
        int val;

        if (!tokens[1]) {
            ph_eprintf("value argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &val) == EINVAL) {
            ph_eprintf("invalid value '%s'", tokens[1]);
            return;
        }
        if (strcmp(tokens[0], CMD_STR_BAR) == 0) {
            osd_value_and_strings(osd_handle, val, tokens + 2, BAR);
        } else {
            osd_value_and_strings(osd_handle, val, tokens + 2, SLIDER);
        }
    } else if (strcmp(tokens[0], CMD_BAR) == 0
               || strcmp(tokens[0], CMD_SLIDER) == 0) {
        int line;
        int val;

        if (!tokens[1]) {
            ph_eprintf("line argument is null");
            return;
        }
        if (!tokens[2]) {
            ph_eprintf("value argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &line) == EINVAL) {
            ph_eprintf("invalid line '%s'", tokens[1]);
            return;
        }
        if (ph_strtoi(tokens[2], &val) == EINVAL) {
            ph_eprintf("invalid value '%s'", tokens[2]);
            return;
        }

        ph_dprintf("line: %d, val: %d", line, val);

        if (strcmp(tokens[0], CMD_BAR) == 0) {
            osd_display(osd_handle, line, val, BAR);
        } else {
            osd_display(osd_handle, line, val, SLIDER);
        }
    } else if (strcmp(tokens[0], CMD_BAR_LENGTH) == 0) {
        int len;

        if (!tokens[1]) {
            ph_eprintf("length argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &len) == EINVAL) {
            ph_eprintf("invalid bar length '%s'", tokens[1]);
        } else {
            set_bar_length(osd_handle, len);
        }
    } else if (strcmp(tokens[0], CMD_COLOR) == 0
               || strcmp(tokens[0], CMD_OUTLINE_COLOR) == 0) {
        if (!tokens[1]) {
            ph_eprintf("color argument is null");
            return;
        }
        if (strcmp(tokens[0], CMD_COLOR) == 0) {
            set_color(osd_handle, tokens[1]);
        } else {
            set_outline_color(osd_handle, tokens[1]);
        }
    } else if (strcmp(tokens[0], CMD_FONT) == 0) {
        if (!tokens[1]) {
            ph_eprintf("font argument is null");
            return;
        }
        set_font(osd_handle, tokens[1]);
    } else if (strcmp(tokens[0], CMD_HIDE) == 0) {
        hide_osd(osd_handle);
    } else if (strcmp(tokens[0], CMD_HORIZONTAL_OFFSET) == 0
               || (strcmp(tokens[0], CMD_OUTLINE_OFFSET) == 0)
               || (strcmp(tokens[0], CMD_SHADOW_OFFSET) == 0)
               || (strcmp(tokens[0], CMD_VERTICAL_OFFSET) == 0)) {
        int offset;

        if (!tokens[1]) {
            ph_eprintf("offset argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &offset) == EINVAL) {
            ph_eprintf("invalid horizontal offset '%s'", tokens[1]);
        } else {
            if (strcmp(tokens[0], CMD_HORIZONTAL_OFFSET) == 0) {
                set_horizontal_offset(osd_handle, offset);
            } else if (strcmp(tokens[0], CMD_OUTLINE_OFFSET) == 0) {
                set_outline_offset(osd_handle, offset);
            } else if (strcmp(tokens[0], CMD_SHADOW_OFFSET) == 0) {
                set_shadow_offset(osd_handle, offset);
            } else {
                set_vertical_offset(osd_handle, offset);
            }
        }
    } else if (strcmp(tokens[0], CMD_POSITION) == 0) {
        if (!tokens[1]) {
            ph_eprintf("position argument is null");
            return;
        } else {
            set_position(osd_handle, tokens[1]);
        }
    } else if (strcmp(tokens[0], CMD_SCROLL) == 0) {
        int lines;

        if (!tokens[1]) {
            ph_eprintf("line count argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &lines) == EINVAL) {
            ph_eprintf("invalid scroll value '%s'", tokens[1]);
        } else {
            scroll_osd(osd_handle, lines);
        }
    } else if (strcmp(tokens[0], CMD_SHOW) == 0) {
        show_osd(osd_handle);
    } else if (strcmp(tokens[0], CMD_STRING) == 0) {
        int line;

        if (!tokens[1]) {
            ph_eprintf("line argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &line) == EINVAL) {
            ph_eprintf("invalid line '%s'", tokens[1]);
            return;
        }
        osd_string(osd_handle, line, tokens[2]);
    } else if (strcmp(tokens[0], CMD_STRINGS) == 0) {
        osd_strings(osd_handle, tokens + 1);
    } else if (strcmp(tokens[0], CMD_TIMEOUT) == 0) {
        int timeout;

        if (!tokens[1]) {
            ph_eprintf("timeout argument is null");
            return;
        }
        if (ph_strtoi(tokens[1], &timeout) == EINVAL) {
            ph_eprintf("invalid timeout '%s'", tokens[1]);
        } else {
            set_timeout(osd_handle, timeout);
        }
    } else {
        ph_eprintf("unknown command '%s'", tokens[0]);
    }
}

/* Initialize daemon. */
static void
init(void)
{
    set_signal_handlers();

    /* Fatal X errors cause program termination via exit(), so clean
     * up fifo in those cases. */
    if (atexit(cleanup_fifo)) {
        ph_eprintf("failed to register exit function");
    }

    osd_handle = init_osd();
    fifo_was_created = make_fifo(arguments.fifo);
}

/* Clean up on regular termination signals. */
void
set_signal_handlers(void)
{
    sigset_t mask;
    int rval;
    struct sigaction sigact, oldact;

    sigemptyset(&mask);
    sigact.sa_handler = termination_handler;
    sigact.sa_mask = mask;
    sigact.sa_flags = 0;
    rval = sigaction(SIGHUP, NULL, &oldact);
    if (rval == 0 && oldact.sa_handler != SIG_IGN) {
        rval = sigaction(SIGHUP, &sigact, NULL) ? -1 : rval;
    }
    rval = sigaction(SIGINT, NULL, &oldact);
    if (rval == 0 && oldact.sa_handler != SIG_IGN) {
        rval = sigaction(SIGINT, &sigact, NULL) ? -1 : rval;
    }
    rval = sigaction(SIGTERM, NULL, &oldact);
    if (rval == 0 && oldact.sa_handler != SIG_IGN) {
        rval = sigaction(SIGTERM, &sigact, NULL) ? -1 : rval;
    }
    if (rval == -1) {
        ph_eprintf("failed to establish signal handlers: %s",
                   strerror(errno));
    }
}

/* Create and return a new OSD object.  Set OSD parameters according
 * to the command line options.
 *
 * If OSD cannot be created, terminate the program. */
osd *
init_osd(void)
{
    osd *osd = create_osd(arguments.nlines);
    if (!osd)_exit(EXIT_FAILURE);

    if (arguments.align) set_align(osd, arguments.align);
    if (arguments.barlen != -1) set_bar_length(osd, arguments.barlen);
    if (arguments.color) set_color(osd, arguments.color);
    if (arguments.outcolor) set_outline_color(osd, arguments.outcolor);
    if (arguments.font) set_font(osd, arguments.font);
    if (arguments.horizoff != -1) set_horizontal_offset(osd, arguments.horizoff);
    if (arguments.outoff != -1) set_outline_offset(osd, arguments.outoff);
    if (arguments.pos) set_position(osd, arguments.pos);
    if (arguments.shadowoff != -1) set_shadow_offset(osd, arguments.shadowoff);
    if (arguments.timeout != -1) set_timeout(osd, arguments.timeout);
    if (arguments.vertoff != -1) set_vertical_offset(osd, arguments.vertoff);

    return osd;
}

/* Create a new named pipe.
 *
 * If fifopath exists, check if it is a named pipe.  If yes, return 0.
 * Otherwise, exit with failure.
 *
 * If fifopath does not exist, try to create it.  In success, return
 * 1, otherwise exit with failure. */
static int
make_fifo(const char *fifopath)
{
    struct stat s;

    assert(fifopath);

    if ((stat(fifopath, &s)) == -1) {
        if (errno != ENOENT) {
            ph_feprintf("can't access file '%s': ", fifopath,
                        strerror(errno));
            exit(EXIT_FAILURE);
        } else {
            /* create a new fifo */
            if (mkfifo(fifopath, S_IRUSR | S_IWUSR) == -1) {
                ph_feprintf("can't open fifo '%s': %s",
                            fifopath, strerror(errno));
                exit(EXIT_FAILURE);
            }
            return 1;
        }
    } else {
        if (!S_ISFIFO(s.st_mode)) {
            ph_feprintf("existing file '%s' is not a named pipe",
                        fifopath);
        }
        return 0;
    }
}

/* Parse program options and arguments. */
static void
parse_options(int argc, char *argv[])
{
    int c;
    extern char *optarg;
    extern int optind, optopt, opterr;

    /* default values for options */
    /* NULL or -1 means unspecified, that is, use OSD default */
    arguments.fifo = NULL;
    arguments.align = NULL;
    arguments.barlen = -1;
    arguments.color = NULL;
    arguments.outcolor = NULL;
    arguments.font = NULL;
    arguments.horizoff = -1;
    arguments.nlines = 2;
    arguments.outoff = -1;
    arguments.pos = NULL;
    arguments.shadowoff = -1;
    arguments.timeout = -1;
    arguments.vertoff = -1;

    while ((c = getopt
            (argc, argv, ":a:b:C:c:f:H:hl:o:p:s:t:V:v")) != -1) {
        switch (c) {
        case 'a':
            arguments.align = optarg;
            break;
        case 'b':
            if (ph_strtoi(optarg, &arguments.barlen) == EINVAL) {
                ph_oeprintf("invalid bar length '%s'", optarg);
            }
            break;
        case 'C':
            arguments.outcolor = optarg;
            break;
        case 'c':
            arguments.color = optarg;
            break;
        case 'f':
            arguments.font = optarg;
            break;
        case 'H':
            if (ph_strtoi(optarg, &arguments.horizoff) == EINVAL) {
                ph_oeprintf("invalid horizontal offset '%s'", optarg);
            }
            break;
        case 'h':
            print_usage();
            break;
        case 'l':
            if (ph_strtoi(optarg, &arguments.nlines) == EINVAL) {
                ph_oeprintf("invalid number of lines '%s'", optarg);
            }
            break;
        case 'o':
            if (ph_strtoi(optarg, &arguments.outoff) == EINVAL) {
                ph_oeprintf("invalid outline offset '%s'", optarg);
            }
            break;
        case 'p':
            arguments.pos = optarg;
            break;
        case 's':
            if (ph_strtoi(optarg, &arguments.shadowoff) == EINVAL) {
                ph_oeprintf("invalid shadow offset '%s'", optarg);
            }
            break;
        case 't':
            if (ph_strtoi(optarg, &arguments.timeout) == EINVAL) {
                ph_oeprintf("invalid timeout '%s'", optarg);
            }
            break;
        case 'V':
            if (ph_strtoi(optarg, &arguments.vertoff) == EINVAL) {
                ph_oeprintf("invalid vertical offset '%s'", optarg);
            }
            break;
        case 'v':
            print_version();
            break;
        case ':':
            ph_oeprintf("option '-%c' requires an argument", optopt);
            break;
        case '?':
            ph_oeprintf("unrecognized option '-%c'", optopt);
            break;
        }
    }

    if (optind == argc) {
        ph_oeprintf("too few arguments");
    } else {
        arguments.fifo = argv[optind];
    }
}

/* Print longer usage message to stdout and exit with success. */
static void
print_usage(void)
{
    printf("Usage: xosdd [OPTION...] FIFO\n"
           "Invoke XOSD daemon for listening requests from FIFO.\n"
           "\n"
           "Options:\n"
           "  -a ALIGN  Use alignment ALIGN; ALIGN can be 'left', 'right', or 'center'\n"
           "  -b N      Set bar length to N percent\n"
           "  -C COLOR  Set outline color to COLOR\n"
           "  -c COLOR  Set foreground color to COLOR\n"
           "  -f FONT   Use font FONT\n"
           "  -H N      Set horizontal offset to N pixels\n"
           "  -l N      Allocate N lines for OSD\n"
           "  -o N      Set outline offset to N pixels\n");
    printf("  -p POS    Set position to POS; POS is either 'top' or 'bottom'\n"
           "  -s N      Set shadow offset to N pixels\n"
           "  -t N      Set timeout to N seconds\n"
           "  -V N      Set vertical offset to N pixels\n"
           "\n"
           "  -h        Give this help list\n"
           "  -v        Print program version\n"
           "\n"
           "Report bugs to <petterih@iki.fi>.\n");

    exit(EXIT_SUCCESS);
}

/* Print version information to stdout and exit with success. */
static void
print_version(void)
{
    printf("xosdd " VERSION "\n"
           COPYRIGHT "\n"
           "This is free software; see the source for copying conditions.  There is NO\n"
           "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
    exit(EXIT_SUCCESS);
}

/* Read one line from fifopath and break it into whitespace-separated
 * tokens for handle_command.  If the command is "exit", stop
 * processing and return. */
static void
process_requests(char *fifopath, osd *osd)
{
    FILE *fifo;
    char *msg;
    char **tokens;
    char done = 0;

    assert(fifopath);
    ph_dprintf("processing requests");

    while (!done) {
        ph_dprintf("opening fifo");
        if ((fifo = fopen(fifopath, "r")) == NULL) {
            ph_eprintf("can't open fifo '%s' for reading",
                       fifopath);
            return;
        }

        while ((msg = ph_getline(fifo)) && !done) {
            if (strlen(msg) == 0) {
                ph_dprintf("got empty message");
                free(msg);
                continue;
            }

            ph_dprintf("got message '%s'", msg);
            tokens = ph_strtok(msg, " ", '"');
            ph_dprintf("command is '%s'", tokens[0]);
            if (tokens[0] == NULL) {
                ph_dprintf("empty command");
            } else if (strcmp(tokens[0], "term") == 0) {
                ph_dprintf("got term command, terminating");
                done = 1;
            } else {
                ph_dprintf("handling command '%s'", tokens[0]);
                handle_command(tokens);
                ph_dprintf("done handling command '%s'", tokens[0]);
            }

            ph_free_tokens(tokens);
            free(msg);
        }

        ph_dprintf("closing fifo");
        fclose(fifo);
    }
}

/* Handler for "normal exit" signals. */
static void
termination_handler(int signum)
{
    if (cleanup_in_progress) raise(signum);
    cleanup_in_progress = 1;
    cleanup_fifo();
    signal(signum, SIG_DFL);
    raise(signum);
}

int
main(int argc, char *argv[])
{
    setlocale(LC_ALL, "");
    ph_set_program_name("xosdd");
    parse_options(argc, argv);
    ph_set_debug_level(0);
    init();
    process_requests(arguments.fifo, osd_handle);
    cleanup_fifo();
    cleanup_osd();
    return 0;
}
